#!/bin/sh
#
# calib.tcl
#
# Restart with wish \
exec wish "$0" "$@"

# Plot samples as dots?
set do_samples 1

set graph_w 500
set graph_h 400

# Don't start calibration until user says so.
set calibrating 0

set mass ""
set highlight ""

font create regular -family Helvetica -size -12
font create bold -family Helvetica -size -12 -weight bold
font create fixed -family lucidatypewriter -size -14 -weight bold

# Input data format:
#
# PRESSURE a b c d     a, b, c, d are ASCII-encoded nonnegative integers below
#                      256.  Together they represent a simultaneous set of
#                      samples read from four pressure sensors.
#
# All else is ignored.

# Converts from mass to X coordinate.
proc mass_to_x {i} {
    global graph_w
    return [expr {($graph_w - 40) / 5.5 * ($i + 0.25) + 30}]
}

# Converts from X coordinate to mass.
proc x_to_mass {i} {
    global graph_w
    return [expr {($i - 30) * 5.5 / ($graph_w - 40) - 0.25}]
}

# Converts from value to Y coordinate.
proc value_to_y {i} {
    global graph_h
    return [expr {($graph_h - 32) / 256. * (256 - $i) + 10}]
}

# Converts from Y coordinate to value.
proc y_to_value {i} {
    global graph_h
    return [expr {((10 - $i) / ($graph_h - 32.) + 1) * 256}]
}

# Handler for incoming text.
proc incoming {chan} {
    if {[eof $chan]} {
        # No more input.
        close $chan
    } else {
        # Packet received.
        if {[scan [gets $chan] "PRESSURE %d %d %d %d" a b c d] == 4 && !(
                $a < 0 || $a > 255 || $b < 0 || $b > 255 |
                $c < 0 || $c > 255 || $d < 0 || $c > 255)} {
            # Valid packet.
            pressure $a $b $c $d
        } else {
            # Ignore.
        }
    }
} 

# Recursive low-pass filter.
# set avg ""

# Handler for PRESSURE data.
proc pressure {a b c d} {
    global mass freq_max freq_mode freq_sum rect_map freq_tbl do_samples
    global calibrating graph_w graph_h

    # Update the most-recent-packet display.
    .pressure configure -text [format "PRESSURE %3d %3d %3d %3d" $a $b $c $d]

    # Recursive low-pass filter.
    # global avg
    # if {$avg eq ""} {set avg $a}
    # set avg [expr {0.8 * $avg + 0.2 * $a}]
    # set a [expr {int($avg)}]

    # If not calibrating, quit now.
    if {!$calibrating} {
        update idletasks
        return
    }

    set x  [mass_to_x [expr {$mass - 0.125}]]
    set x2 [mass_to_x [expr {$mass + 0.125}]]

    # One more sample...
    incr freq_sum
    set goal [expr {int($freq_sum / 6)}]

    # Update the sample display.
    if {$do_samples} {
        set y [value_to_y [expr {$a + 0.5}]]
        .graph move sample -1 0
        set tag [.graph create oval $graph_w $y $graph_w $y -outline green\
                 -tags sample]
        foreach tag [.graph find overlapping 0 0 30 $graph_h] {
            if {[.graph gettags $tag] eq "sample"} {
                .graph delete $tag
            }
        }
    }

    # Ensure that the rect exists.
    if {[llength [.graph find withtag "vx$mass && vy$a"]] == 0} {
        set y  [value_to_y        $a      ]
        set y2 [value_to_y [expr {$a + 1}]]
        set tag [.graph create rect $x $y $x2 $y2 -tags "vx$mass vy$a val"]
        if {[catch {.graph raise $tag "min || mid || max || mode"}]} {
            .graph lower $tag
        }
        set rect_map($tag) $a
    }

    # Readjust all colors in the column.
    foreach tag [.graph find withtag vx$mass] {
        set pos $rect_map($tag)

        if {$pos == $a} {
            set freq [incr freq_tbl($pos)]
            if {$freq > $freq_max} {
                set freq_max $freq
                set freq_mode $pos
            }
        } else {
            set freq $freq_tbl($pos)
        }

        set val [expr {int($freq * 255. / $freq_max)}]
        set color [format "#%02x%02x%02x" $val $val $val]

        .graph itemconfigure $tag -fill $color -outline $color
    }

    # Update the minimum line.
    set x  [mass_to_x [expr {$mass - 0.25}]]
    set x2 [mass_to_x [expr {$mass + 0.25}]]
    for {set min 0; set tot 0} {$min < 255} {incr min} {
        incr tot $freq_tbl($min)
        if {$tot >= $goal} {
            set tag [.graph find withtag "bx$mass && min"]
            if {[llength $tag] == 0} {
                set tag [.graph create line 0 0 0 0 -fill yellow\
                         -tags "bx$mass min"]
                .graph lower $tag val
            }
            set y [expr {[value_to_y [expr ($min + 1)]] - 1}]
            .graph coords $tag $x $y $x2 $y
            break
        } 
    }

    # Maximum line.
    for {set max 255; set tot 0} {$max >= 0} {incr max -1} {
        incr tot $freq_tbl($max)
        if {$tot >= $goal} {
            set tag [.graph find withtag "bx$mass && max"]
            if {[llength $tag] == 0} {
                set tag [.graph create line 0 0 0 0 -fill yellow\
                         -tags "bx$mass max"]
                .graph lower $tag val
            }
            set y [value_to_y $max]
            .graph coords $tag $x $y $x2 $y
            break
        } 
    }

    # Middle (average) line.
    set tag [.graph find withtag "bx$mass && mid"]
    if {[llength $tag] == 0} {
        set tag [.graph create line 0 0 0 0 -fill blue -tags "bx$mass mid"]
        .graph lower $tag val
    }
    set y [value_to_y [expr {($min + $max) * 0.5}]]
    .graph coords $tag $x $y $x2 $y

    # Mode line.
    set tag [.graph find withtag "bx$mass && mode"]
    if {[llength $tag] == 0} {
        set tag [.graph create line 0 0 0 0 -fill red -tags "bx$mass mode"]
        .graph lower $tag val
    }
    set y [expr {[value_to_y $freq_mode] + 1}]
    .graph coords $tag $x $y $x2 $y

    update idletasks
}

# Handler for mouseover, motion, and mouseout for graph canvas.
proc motion_graph {x y} {
    global graph_w graph_h mass highlight

    # Handle horizontal ruler.
    set i [y_to_value $y]
    if {$i >= 0 && $i <= 256} {
        # Create/update horizontal ruler.
        set tag [.graph find withtag yline]
        if {[llength $tag] == 0} {
            set tag [.graph create line 0 0 0 0 -fill lightblue -dash "4 4"\
                    -tags yline]
        }
        .graph coords $tag 27 $y [expr {$graph_w - 5}] $y

        # Create/update text outline.
        set tag [.graph find withtag ybox]
        if {[llength $tag] == 0} {
            set tag [.graph create rectangle 0 0 0 0 -outline lightblue\
                    -fill black -dash "4 4" -tags ybox]
        }
        .graph coords $tag 0 [expr {$y - 10}] 27 [expr {$y + 8}]

        # Create/update text label.
        set tag [.graph find withtag ytext]
        if {[llength $tag] == 0} {
            set tag [.graph create text 0 0 -fill white -anchor e -tags ytext]
        }
        .graph itemconfigure $tag -text [expr {int($i)}]
        .graph coords $tag 24 $y
    } else {
        # Outside bounds.
        foreach tag [.graph find withtag "yline || ybox || ytext"] {
            .graph delete $tag
        }
    }

    # Handle vertical ruler.
    set i [expr {floor(([x_to_mass $x] + 0.125) / 0.25) * 0.25}]
    if {$i < 0.0 || $i > 5.0 || $i == $mass} {set i ""}

    if {$i != $highlight} {
        if {$highlight ne "" && $highlight != $mass} {
            # Erase old highlight.
            .graph itemconfigure [.graph find withtag gx$highlight] -fill gray20
            .graph delete xbox
            .graph delete xtext
        } else {
            # Nothing to do.
        }

        if {$i ne ""} {
            # Create new highlight.
            set xc [mass_to_x $i]
            set x1 [expr {$xc - 20}]
            set y1 [expr {$graph_h - 18}]
            set x2 [expr {$xc + 20}]
            set y2 [expr {$graph_h -  2}]

            # Vertical line.
            .graph itemconfigure [.graph find withtag gx$i] -fill lightblue

            # Text outline.
            .graph create rectangle $x1 $y1 $x2 $y2 -tags xbox -fill black\
                   -outline lightblue -dash "4 4"

            # Text label.
            .graph create text $xc [expr {$y1 + 2}] -tags xtext -fill white\
                   -anchor n -text $i -font bold
        } else {
            # Nothing to do.
        }

        set highlight $i
    }
}

# Handler for clicks on the graph.
proc select_graph {x y} {
    global highlight

    if {$highlight ne ""} {
        set hl $highlight
        motion_graph -1 $y
        select_mass $hl
    }
}

# Selects a mass to calibrate.
proc select_mass {val} {
    global calibrating mass freq_max freq_mode freq_sum freq_tbl graph_h

    if {$mass == $val} {return}

    set calibrating 0 

    if {$mass ne ""} {
        # Kill old calibration line.
        .graph delete mbox
        .graph delete mtext
        .graph itemconfigure [.graph find withtag gx$mass] -fill gray20 -width 1
    }

    # Reset calibration information for the new mass.
    set mass $val
    set freq_max  0
    set freq_mode 0
    set freq_sum  0
    for {set i 0} {$i < 256} {incr i} {
        set freq_tbl($i) 0
    }
    foreach tag [.graph find withtag vx$mass] {
        .graph delete $tag
    }

    # Draw a new calibration line.
    set xc [mass_to_x $val]
    set x1 [expr {$xc - 20}]
    set y1 [expr {$graph_h - 18}]
    set x2 [expr {$xc + 20}]
    set y2 [expr {$graph_h -  2}]

    # Vertical line.
    .graph itemconfigure [.graph find withtag gx$val] -fill orange -width 2

    # Text outline.
    .graph create rectangle $x1 $y1 $x2 $y2 -tags mbox -fill black\
           -outline orange -dash "4 4"

    # Text label.
    .graph create text $xc [expr {$y1 + 2}] -tags mtext -fill white\
           -anchor n -text $val -font bold
}

proc main {} {
    global graph_w graph_h

    # Create widgets.
    frame  .border1    -relief flat -borderwidth 2
    frame  .border2    -relief sunken -borderwidth 2
    canvas .graph      -width $graph_w -height $graph_h -background black\
                       -highlightthickness 0
    label  .pressure   -text "No pressure data received" -font fixed
    checkbutton .calib -text Calibrating -font bold -variable calibrating

    bind .graph <Enter>    "motion_graph %x %y"
    bind .graph <Motion>   "motion_graph %x %y"
    bind .graph <Leave>    "motion_graph -1 -1"
    bind .graph <Button-1> "select_graph %x %y"

    # Draw horizontal gridlines and labels.
    for {set i 0} {$i <= 256} {incr i 16} {
        set y [value_to_y $i]

        # Create label.
        .graph create text 24 $y -text $i -fill gray60 -anchor e -font regular\
               -tags "ly$i ylabel"

        # Create line.
        set tag [.graph create line 27 $y [expr {$graph_w - 5}] $y]

        if {$i == 0} {
            # Axis.
            .graph itemconfigure $tag -fill gray80 -width 2 -tags axis
        } else {
            # Gridline.
            .graph itemconfigure $tag -fill gray20 -dash "2 4"\
                   -tags "gy$i ygrid"
        }
    }

    # Draw vertical gridlines and labels.
    for {set i -0.25} {$i <= 5.25} {set i [expr {$i + 0.25}]} {
        set x [mass_to_x $i]
        set y [expr {$graph_h - 16}]

        if {int($i * 2) == $i * 2 && $i >= 0.0} {
            # Label.
            .graph create text $x $y -text [format "%.1f" $i] -fill gray60\
                   -anchor n -tags "lx$i xlabel" -font regular
        }

        # Line.
        set tag [.graph create line $x 5 $x [expr {$graph_h - 18}]]

        if {$i == -0.25} {
            # Axis.
            .graph itemconfigure $tag -fill gray80 -width 2 -tags axis
        } else {
            # Gridline.
            .graph itemconfigure $tag -fill gray20 -dash "2 4"\
                   -tags "gx$i xgrid"
        }
    }

    .graph raise axis

    # Pack widgets into window.
    pack .graph    -fill both -in .border2
    pack .border2  -fill both -in .border1
    pack .pressure -fill x    -in .border1 -side left -expand 1
    pack .calib               -in .border1 -side left
    pack .border1  -fill both

    # Default mass.
    select_mass 2.5

    # Calculate and fix the window size.
    wm resizable . 0 0
    update
    wm geometry . [wm geometry .]

    # Arrange for incoming data to be processed.
    fileevent stdin readable "incoming stdin"
}

main

# vim: set ts=4 sts=4 sw=4 tw=80 et:
